#include "stdafx.h"
#include "newtrick.h"
#include "image.h"
#include "delta.h"

DeltaControl::DeltaControl() {
	static struct deltacontrol d = {0,0,0,0};
	m_Major = d;
	m_Minor = d;
	m_ErrorTerm = 0;
}

void DeltaControl::SetFrom(int Major, int Minor) {
	m_Major.current = Major;
	m_Minor.current = Minor;
}

void DeltaControl::SetTo(int Major, int Minor) {
	SetTo(Major, &m_Major);
	SetTo(Minor, &m_Minor);
	m_ErrorTerm = m_Major.len / 2;
	m_MinorLast = TRUE;
}

BOOL DeltaControl::Next() {
	if (m_MinorLast) m_MinorLast = FALSE;
	else {
		m_ErrorTerm += m_Minor.len;
		if (m_ErrorTerm > m_Major.len && m_Minor.togo) {
			m_Minor.current += m_Minor.delta;
			OnMinorChange(m_Minor.current);
			m_ErrorTerm -= m_Major.len;
			m_Minor.togo--;
			m_MinorLast = TRUE;
			return TRUE;
		}
	}
	if (!m_Major.togo) return FALSE;
	m_Major.current += m_Major.delta;
	m_Major.togo--;
	OnMajorChange(m_Major.current);
	return TRUE;
}

void DeltaControl::SetTo(int In, deltacontrol *Out) {
	int i = In - Out->current;

	Out->delta = (i < 0) ? -1 : 1;
	Out->togo = Out->len = abs(i);
}

DeltaControlLine::DeltaControlLine(CPoint *From, CPoint *To) {
	Init(From, To);
}

DeltaControlLine::DeltaControlLine(CPoint *From, CPoint *To, CPoint *Offset) {
	Init(From, To);
	CPoint::Offset(*Offset);
	m_Major.current = *m_PMajor;
	m_Minor.current = *m_PMinor;
}

void DeltaControlLine::Init(CPoint *From, CPoint *To) {
	CPoint p = *From;
	p -= *To;
	if (abs(p.x) > abs(p.y)) {
		SetFrom(From->x, From->y);
		SetTo(To->x, To->y);
		m_PMajor = &x;
		m_PMinor = &y;
	}
	else {
		SetFrom(From->y, From->x);
		SetTo(To->y, To->x);
		m_PMajor = &y;
		m_PMinor = &x;
	}
	x = From->x;
	y = From->y;
}

DeltaControlPair::DeltaControlPair(DeltaControl *A, DeltaControl *B) {
	SetFrom(0, 0);
	if (A->GetLength() > B->GetLength()) {
		SetTo(A->GetLength(), B->GetLength());
		m_PMajor = A;
		m_PMinor = B;
	}
	else {
		SetTo(B->GetLength(), A->GetLength());
		m_PMajor = B;
		m_PMinor = A;
	}
}

BOOL DeltaControlPair::Next() {
	m_ErrorTerm += m_Minor.len;
	if (m_ErrorTerm > m_Major.len && m_Minor.togo) {
		m_Minor.current += m_Minor.delta;
		OnMinorChange(m_Minor.current);
		m_ErrorTerm -= m_Major.len;
		m_Minor.togo--;
	}
	if (!m_Major.togo) return FALSE;
	m_Major.current += m_Major.delta;
	m_Major.togo--;
	OnMajorChange(m_Major.current);
	return TRUE;
}

