#include "stdafx.h"
#include "newtrick.h"
#include "image.h"
#include "delta.h"

class CIBHelper : public CObject
{
public:
	CIBHelper(CImageBitmap *Parent = 0);
	virtual ~CIBHelper();
	virtual void LoadDataFromBMPFile(BITMAPINFOHEADER *Hdr, char *Image) {};
	virtual void *GetPixel(CPoint *Point) {return 0;};
	virtual BOOL PutPixel(CPoint *Point, void *In) {return 0;};
	BOOL GoodCoords(CPoint *Point);
	operator char *() {return m_Image;};
	operator int() {return m_ImageSize;};
protected:
	CImageBitmap *m_Parent;
	char *GetBitmapBits();
	char **GetBitmapRows();
	char *m_Image;
	int m_ImageSize;
	char **m_Rows;
};

class CIBHelper32 : public CIBHelper
{
public:
	CIBHelper32(CImageBitmap *Parent = 0);
	virtual ~CIBHelper32() {};
	virtual void LoadDataFromBMPFile(BITMAPINFOHEADER *Hdr, char *Image);
	virtual void *GetPixel(CPoint *Point);
	virtual BOOL PutPixel(CPoint *Point, void *In);
};

class CIBHelper24 : public CIBHelper
{
public:
	CIBHelper24(CImageBitmap *Parent = 0);
	virtual ~CIBHelper24() {};
	virtual void LoadDataFromBMPFile(BITMAPINFOHEADER *Hdr, char *Image);
	virtual void *GetPixel(CPoint *Point);
	virtual BOOL PutPixel(CPoint *Point, void *In);
};

class CIBLine : public DeltaControlLine
{
public:
	CIBLine(CPoint *From, CPoint *To, CPoint *Input, CIBHelper *Source, CPoint *Offset, CIBHelper *Dest);
	virtual ~CIBLine() {};
protected:
	virtual void OnMajorChange(int Current);
	virtual void OnMinorChange(int Current);
private:
	void PutPixel();
	CPoint *m_Input;
	CIBHelper *m_Source;
	CIBHelper *m_Dest;
};

CImageBitmap::CImageBitmap() {
	m_Private = 0;
	m_Image = 0;
}

CImageBitmap::~CImageBitmap() {
	if (m_Private) delete m_Private;
	m_Private = 0;
	if (m_Image) delete m_Image;
	m_Image = 0;
}

#define READ(P, S) if (f.Read(P, S) != S) break;

BOOL CImageBitmap::LoadBMPFile(const char *Filespec) {
	BOOL ret = FALSE;
	CFile f;

	m_Title = Filespec;
	if (m_Image) delete m_Image;
	m_Image = 0;
	if (!f.Open(Filespec, CFile::modeRead)) return FALSE;
	while (!ret) {
		BITMAPFILEHEADER bmfh;
		READ(&bmfh, sizeof(BITMAPFILEHEADER));
		READ(&m_Hdr, sizeof(BITMAPINFOHEADER));
		f.Seek(bmfh.bfOffBits, CFile::begin);
		if (!m_Hdr.biSizeImage)
			m_Hdr.biSizeImage = bmfh.bfSize - bmfh.bfOffBits;
		m_Size.cx = m_Hdr.biWidth;
		m_Size.cy = m_Hdr.biHeight;
		m_Image = new char[m_Hdr.biSizeImage];
		while (m_Image) {
			READ(m_Image, m_Hdr.biSizeImage);
			ret = TRUE;
			break;
		}
		break;
	}
	f.Close();
	if (m_Image && !ret) {
		delete m_Image;
		m_Image = 0;
	}
	return ret;
}

BOOL CImageBitmap::CreateFromBMPFile(CDC *pDC) {
	if (!m_Image) return FALSE;
	if (!CreateCompatibleBitmap(pDC, m_Hdr.biWidth, m_Hdr.biHeight))
		return FALSE;
	if (!GetBitmap(&m_Bitmap)) return FALSE;
	switch (m_Bitmap.bmBitsPixel) {
		case 32: m_Private = new CIBHelper32(this);break;
		case 24: m_Private = new CIBHelper24(this);break;
		default: CString s;
			s.Format("Unsupported display bitdepth: %d bpp",
				m_Bitmap.bmBitsPixel);
			AfxMessageBox(s);
			return FALSE;
	}
	LoadDataFromBMPFile();
	return TRUE;
}

void CImageBitmap::LoadDataFromBMPFile() {
	if (m_Private) m_Private->LoadDataFromBMPFile(&m_Hdr, m_Image);
}

void *CImageBitmap::GetPixel(CPoint *Point) {
	if (m_Private) return m_Private->GetPixel(Point);
	return 0;
}

BOOL CImageBitmap::PutPixel(CPoint *Point, void *In) {
	if (m_Private) return m_Private->PutPixel(Point, In);
	return 0;
}

void CImageBitmap::Transform(CImageBitmap *In, CPoint *TopLeft,   // ) {
		CPoint *BottomLeft, CPoint *TopRight, CPoint *BottomRight,
		CPoint *Offset) {
	CIBHelper *p = *In;
	BITMAP *b = *In;

	// controllers for moving the end points of the line to be stretched
	DeltaControlLine left(TopLeft, BottomLeft);
	DeltaControlLine right(TopRight, BottomRight);
	DeltaControlPair output_y(&left, &right);

	// controller for moving down the input Y direction
	DeltaControl input_y; input_y.SetTo(b->bmHeight, 0);
	DeltaControlPair process_y(&input_y, &output_y);

	do {
		// controller for outputing pixels in X direction
		DeltaControl input_x; input_x.SetTo(b->bmWidth, 0);
		CPoint pt(0, input_y.GetMajor());
		CIBLine output_x(&left, &right, &pt, p, Offset,
			*this);
		DeltaControlPair process_x(&input_x, &output_x);

		do { pt.x = input_x.GetMajor();
		} while (process_x.Next());
	} while (process_y.Next());

	SetBitmapBits(*m_Private, *m_Private);
}

CIBHelper::CIBHelper(CImageBitmap *Parent) {
	m_Parent = Parent;
	m_Image = 0;
	m_Rows = 0;
}

CIBHelper::~CIBHelper() {
	if (m_Image) delete m_Image;
	m_Image = 0;
	if (m_Rows) delete m_Rows;
	m_Rows = 0;
}

char *CIBHelper::GetBitmapBits() {
	BITMAP *b = *m_Parent;
	m_ImageSize = b->bmHeight * b->bmWidthBytes;
	if (m_Image) delete m_Image;
	if (m_Rows) delete m_Rows;
	m_Rows = 0;
	m_Image = new char [m_ImageSize];
	m_Parent->GetBitmapBits(m_ImageSize, m_Image);
	return m_Image;
}

char **CIBHelper::GetBitmapRows() {
	if (!GetBitmapBits()) return 0;
	BITMAP *b = *m_Parent;
	int y;

	m_Rows = new char *[b->bmHeight];
	char *c = m_Image;
	for (y = 0; y < b->bmHeight; y++, c += b->bmWidthBytes) m_Rows[y] = c;
	return m_Rows;
}

BOOL CIBHelper::GoodCoords(CPoint *Point) {
	if (Point->x < 0 || Point->y < 0) return FALSE;
	BITMAP *b = *m_Parent;
	if (Point->x < b->bmWidth && Point->y < b->bmHeight) return TRUE;
	return FALSE;
}

CIBHelper32::CIBHelper32(CImageBitmap *Parent) : CIBHelper(Parent){}

void CIBHelper32::LoadDataFromBMPFile(BITMAPINFOHEADER *Hdr, char *Image) {
	BITMAP *b = *m_Parent;
	char **c = GetBitmapRows();
	int i, x, y, z;

	// compute 32 bit boundary requirement
	i = ((1 + Hdr->biWidth) * 3) & ~3;
	for (y = 0, z = b->bmHeight; y < b->bmHeight && y < Hdr->biHeight &&
			c && Hdr->biBitCount == 24; y++) {
		RGBTRIPLE *r = (RGBTRIPLE *)Image;
		z--;
		RGBQUAD *q = (RGBQUAD *)c[z];
		for (x = 0; x < b->bmWidth && x < Hdr->biWidth;
				x++, r++, q++) {
			q->rgbBlue = r->rgbtBlue;
			q->rgbGreen = r->rgbtGreen;
			q->rgbRed = r->rgbtRed;
			q->rgbReserved = 0;
		}
		// move pointer to next row
		Image += i;
	}
	m_Parent->SetBitmapBits(m_ImageSize, m_Image);
}

void *CIBHelper32::GetPixel(CPoint *Point) {
	if (!GoodCoords(Point)) return 0;
	if (!m_Rows) {
		if (!GetBitmapRows()) return 0;
	}
	int *p = (int *)m_Rows[Point->y];
	return &p[Point->x];
}

BOOL CIBHelper32::PutPixel(CPoint *Point, void *In) {
	if (!In) return FALSE;
	if (!GoodCoords(Point)) return FALSE;
	if (!m_Rows) {
		if (!GetBitmapRows()) return FALSE;
	}
	int *i = (int *)In;
	int *p = (int *)m_Rows[Point->y];
	p[Point->x] = *i;
	return TRUE;
}

CIBHelper24::CIBHelper24(CImageBitmap *Parent) : CIBHelper(Parent){}

void CIBHelper24::LoadDataFromBMPFile(BITMAPINFOHEADER *Hdr, char *Image) {
	BITMAP *b = *m_Parent;
	char **c = GetBitmapRows();
	int i, x, y, z;

	// compute 32 bit boundary requirement
	i = ((1 + Hdr->biWidth) * 3) & ~3;
	for (y = 0, z = b->bmHeight; y < b->bmHeight && y < Hdr->biHeight &&
			c && Hdr->biBitCount == 24; y++) {
		RGBTRIPLE *r = (RGBTRIPLE *)Image;
		z--;
		RGBTRIPLE *q = (RGBTRIPLE *)c[z];
		for (x = 0; x < b->bmWidth && x < Hdr->biWidth;
				x++, r++, q++) *q = *r;
		// move pointer to next row
		Image += i;
	}
	m_Parent->SetBitmapBits(m_ImageSize, m_Image);
}

void *CIBHelper24::GetPixel(CPoint *Point) {
	if (!GoodCoords(Point)) return 0;
	if (!m_Rows) {
		if (!GetBitmapRows()) return 0;
	}
	RGBTRIPLE *p = (RGBTRIPLE *)m_Rows[Point->y];
	return &p[Point->x];
}

BOOL CIBHelper24::PutPixel(CPoint *Point, void *In) {
	if (!In) return FALSE;
	if (!GoodCoords(Point)) return FALSE;
	if (!m_Rows) {
		if (!GetBitmapRows()) return FALSE;
	}
	RGBTRIPLE *i = (RGBTRIPLE *)In;
	RGBTRIPLE *p = (RGBTRIPLE *)m_Rows[Point->y];
	p[Point->x] = *i;
	return TRUE;
}

CIBLine::CIBLine(CPoint *From, CPoint *To, CPoint *Input,   // ) {
		CIBHelper *Source, CPoint *Offset, CIBHelper *Dest)
		: DeltaControlLine(From, To, Offset) {
	m_Input = Input;
	m_Source = Source;
	m_Dest = Dest;
	// output the first pixel
	PutPixel();
}

void CIBLine::OnMajorChange(int Current) {
	DeltaControlLine::OnMajorChange(Current);
	PutPixel();
}

void CIBLine::OnMinorChange(int Current) {
	DeltaControlLine::OnMinorChange(Current);
	PutPixel();
}

void CIBLine::PutPixel() {
	m_Dest->PutPixel(this, m_Source->GetPixel(m_Input));
}


