#include "stdafx.h"
#include "wndszpos.h"

void dbWndSize::OnInitDialog(CWnd *Parent, CWnd *Control) {
	CRect r, s;

	m_Wnd = Control;
	m_Wnd->GetWindowRect(r);
	Parent->GetWindowRect(s);
	m_TopLeft = r.TopLeft() - s.TopLeft();
	m_BottomRight = s.BottomRight() - r.BottomRight();
	CSize c(1, 1);
	m_BottomRight += c;
}

void dbWndSize::OnInitDialog(CDialog *Parent, CWnd *Control) {
	CRect r, s;

	OnInitDialog((CWnd *)Parent, Control);
	Parent->GetClientRect(r);
	Parent->GetWindowRect(s);
	CSize c((s.Width() - r.Width()), (s.Height() - r.Height()));
	m_BottomRight -= c;
}

void dbWndSize::OnSize(UINT nType, int cx, int cy) {
	if (!m_Wnd) return;
	switch (nType) {
		case SIZE_MINIMIZED: return;
	}
	cx -= m_BottomRight.x; if (cx < m_TopLeft.x) return;
	cx -= m_TopLeft.x; if (cx < 0) return;
	cy -= m_BottomRight.y; if (cy < m_TopLeft.y) return;
	cy -= m_TopLeft.y; if (cy < 0) return;
	m_Wnd->SetWindowPos(0, 0, 0, cx, cy, SWP_NOMOVE | SWP_NOZORDER);
}

void dbWndPosition::OnInitDialog(CWnd *Parent, CWnd *Control, int Minx,   // ) {
		int Miny) {
	CRect r, s;

	m_Wnd = Control;
	m_Wnd->GetWindowRect(r);
	Parent->GetWindowRect(s);
	m_TopLeft = r.TopLeft() - s.TopLeft();
	m_BottomRight = s.BottomRight() - s.TopLeft();
	m_MinX = (Minx == -1) ? m_TopLeft.x : Minx;
	m_MinY = (Miny == -1) ? m_TopLeft.y : Miny;
}

void dbWndPosition::OnSize(UINT nType, int cx, int cy) {
	if (!m_Wnd) return;
	switch (nType) {
		case SIZE_MINIMIZED: return;
	}
	if (cx < m_MinX || cy < m_MinY) return;
	cx -= m_BottomRight.x; cx += m_TopLeft.x; if (cx < 0) return;
	cy -= m_BottomRight.y; cy += m_TopLeft.y; if (cy < 0) return;
	m_Wnd->SetWindowPos(0, cx, cy, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}

void dbDlgSize::SetBottom(CDialog *Dialog, CWnd *Original, CWnd *New) {
	int i;
	CRect r, s;

	New->GetWindowRect(r);
	Original->GetWindowRect(s);
	i = s.bottom - r.bottom;
	Dialog->GetWindowRect(s);
	Dialog->SetWindowPos(0, 0, 0, s.Width(), s.Height() - i,
		SWP_NOMOVE | SWP_NOZORDER);
}

void dbDlgSize::SetRight(CDialog *Dialog, CWnd *Original, CWnd *New) {
	int i;
	CRect r, s;

	New->GetWindowRect(r);
	Original->GetWindowRect(s);
	i = s.right - r.right;
	Dialog->GetWindowRect(s);
	Dialog->SetWindowPos(0, 0, 0, s.Width() - 1, s.Height(),
		SWP_NOMOVE | SWP_NOZORDER);
}

class dbDlgRowCntrl : public CObject
{
public:
	dbDlgRowCntrl(CWnd *Control, int MinWidth, int MaxWidth);
	virtual ~dbDlgRowCntrl() {};
	operator CRect *() {return &m_Rect;}
	int GetMinWidth() {return m_MinWidth;}
	int GetMaxWidth() {return m_MaxWidth;}
	int GetCurrentWidth() {return m_CurrentWidth;}
	BOOL IsResizeable() {return (m_MinWidth != m_MaxWidth);}
	int UpSize(int Delta);
	int DownSize(int Delta);
	void SetPos(int *X, int Y, int Spacing);
private:
	CWnd *m_Parent;
	CRect m_Rect;
	int m_MinWidth;
	int m_MaxWidth;
	int m_CurrentWidth;
};

void dbDlgRow::OnInitDialog(CDialog *Parent, int MinXSpacing) {
	m_Parent = Parent;
	m_MinXSpacing = MinXSpacing;
	m_FixedWidth = 0;
}

void dbDlgRow::OnInitDialog(CWnd *Control, int MinWidth, int MaxWidth) {
	dbDlgRowCntrl *d =new dbDlgRowCntrl(Control, MinWidth, MaxWidth);
	m_Controls.Add(d);
	if (d->IsResizeable()) m_Resizeable.Add(d);
	else m_FixedWidth += d->GetMinWidth();
}

void dbDlgRow::OnInitDialog(int MinYPos, int MinXSize) {
	CRect r, *s, *t;
	dbDlgRowCntrl *d = m_Controls.GetFirst();

	m_MinYPos = MinYPos;
	m_MinXSize = MinXSize;
	m_LastWidth = m_LastSpacing = -1;
	m_BottomSpacing = -1;
	if (!d || !m_Parent) return; 
	m_Parent->GetWindowRect(r);
	s = *d;
	d = m_Controls.GetLast();
	t = *d;
	m_TopLeft = s->TopLeft() - r.TopLeft();
	m_BottomRight = t->BottomRight() - r.TopLeft();
	m_Valid = TRUE;
	if (MinXSize > 1) return;
	m_MinXSize = (m_TopLeft.x * 2) +
		(m_MinXSpacing * m_Controls.GetUpperBound());
	CA_ITERATE(m_Controls, dbDlgRowCntrl, m_MinXSize += c->GetMinWidth());
}

void dbDlgRow::OnInitDialog() {
	CRect r, *s, t;
	dbDlgRowCntrl *d = m_Controls.GetFirst();

	m_MinYPos = m_MinXSize = m_LastWidth = m_LastSpacing = -1;
	if (!d || !m_Parent) return; 
	m_Parent->GetWindowRect(r);
	m_Parent->GetClientRect(t);
	s = *d;
	m_TopLeft = s->TopLeft() - r.TopLeft();
	m_BottomSpacing = t.Height() - (s->top - r.top);
	m_Valid = TRUE;
	m_MinXSize = (m_TopLeft.x * 2) +
		(m_MinXSpacing * m_Controls.GetUpperBound());
	CA_ITERATE(m_Controls, dbDlgRowCntrl, m_MinXSize += c->GetMinWidth());
}

void dbDlgRow::OnSize(UINT nType, int cx, int cy) {
	if (!m_Valid) return;
	switch (nType) {
		case SIZE_MINIMIZED: return;
	}
	if (cx < m_MinXSize || cy < m_MinYPos) return;
	if (m_BottomSpacing > 0) cy -= m_BottomSpacing;
	else {	cy -= m_BottomRight.y; cy += m_TopLeft.y; }
	if (cy < 0) return;
	cx -= m_TopLeft.x;
	cx -= m_TopLeft.x;
	while (cx != m_LastWidth) {
		m_LastWidth = cx;
		int i = 0;
		CA_ITERATE(m_Controls, dbDlgRowCntrl, i+= c->GetCurrentWidth());
		int j = (cx - i);
		int k = j / m_Controls.GetUpperBound();
		if (k == m_LastSpacing) break;
		if (m_Resizeable.IsEmpty()) m_LastSpacing = k;
		else Resize(j, k);
		break;
	}
	cx = m_TopLeft.x;
	CA_ITERATE(m_Controls, dbDlgRowCntrl, c->SetPos(&cx,cy,m_LastSpacing));
}

#define UPSIZE {Delta = c->UpSize(Delta); if (!Delta) break;}
#define DOWNSIZE {Delta = c->DownSize(Delta); if (!Delta) break;}

void dbDlgRow::Resize(int Delta, int Spacing) {
	Delta -= m_LastSpacing * m_Controls.GetSize();
	if (Spacing < m_LastSpacing) {
		if (Spacing >= m_MinXSpacing) {
			m_LastSpacing = Spacing;
			return;
		}
		CA_ITERATE(m_Resizeable, dbDlgRowCntrl, DOWNSIZE);
	}
	else	CA_ITERATE(m_Resizeable, dbDlgRowCntrl, UPSIZE);
	m_LastSpacing = Spacing + (Delta / m_Controls.GetUpperBound());
}

dbDlgRowCntrl::dbDlgRowCntrl(CWnd *Control, int MinWidth, int MaxWidth) {
	m_Parent = Control;
	m_Parent->GetWindowRect(&m_Rect);
	m_CurrentWidth = m_Rect.Width();
	if (MaxWidth == -1) m_MaxWidth = m_CurrentWidth;
	else m_MaxWidth = MaxWidth;
	if (MinWidth == -1) m_MinWidth = m_CurrentWidth;
	else m_MinWidth = MinWidth;
}

int dbDlgRowCntrl::UpSize(int Delta) {
	int i = m_CurrentWidth + Delta;

	if (i > m_MaxWidth) i = m_MaxWidth;
	Delta -= (i - m_CurrentWidth);
	m_CurrentWidth = i;
	return Delta;
}

int dbDlgRowCntrl::DownSize(int Delta) {
	int i = m_CurrentWidth - Delta;

	if (i < m_MinWidth) i = m_MinWidth;
	Delta -= (m_CurrentWidth - i);
	m_CurrentWidth = i;
	return Delta;
}

void dbDlgRowCntrl::SetPos(int *X, int Y, int Spacing) {
	m_Parent->SetWindowPos(0, *X, Y - m_Rect.Height(), m_CurrentWidth,
		m_Rect.Height(), SWP_NOZORDER);
	*X = *X + Spacing + m_CurrentWidth;
}


