/*******************************************************************************
  File: XDRStream_reader.cpp
  Copyright (c) 1999 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, to prevent the 
  continued distribution of this Software under an open source distribution
  model. This includes, but is not limited to, such things as copyrighting
  or claiming authorship of the Software.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
$Header:$
  Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
  Implements the iXDR_Stream.
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/
// My headers
#include "XDRStream.h"

// Utilities
#include "E2A.h"

// Standard C++ headers
#include <algorithm>

namespace Util {

//
// iXDR_Stream
//
iXDR_Stream::iXDR_Stream()
{
	// empty
}


iXDR_Stream::iXDR_Stream(XDR_Streambuf* sb)
{
	init(sb);
}


iXDR_Stream::~iXDR_Stream()
{
	// empty
}


iXDR_Stream& 
iXDR_Stream::operator>>(bool& n)
{
	sentry qx(*this);
	if (not qx) return *this;
	
	long x;
	*this >> x;
	if (good()) n = (x == 1) ? true : false;
	return *this;
}


iXDR_Stream& 
iXDR_Stream::operator>>(char& n)
{
	sentry qx(*this);
	if (not qx) return *this;
	
	long x;
	*this >> x;
	if (good()) n = static_cast<signed char>(x);
	return *this;
}


iXDR_Stream& 
iXDR_Stream::operator>>(signed char& n)
{
	sentry qx(*this);
	if (not qx) return *this;

	long x;
	*this >> x;
	if (good()) n = static_cast<signed char>(x);
	return *this;
}


iXDR_Stream& 
iXDR_Stream::operator>>(unsigned char& n)
{
	sentry qx(*this);
	if (not qx) return *this;

	unsigned long x;
	*this >> x;
	if (good()) n = static_cast<unsigned char>(x);
	return *this;
}


iXDR_Stream& 
iXDR_Stream::operator>>(short& n)
{
	sentry qx(*this);
	if (not qx) return *this;

	long x;
	*this >> x;
	if (good()) n = static_cast<short>(x);
	return *this;
}


iXDR_Stream& 
iXDR_Stream::operator>>(unsigned short& n)
{
	sentry qx(*this);
	if (not qx) return *this;

	unsigned long x;
	*this >> x;
	if (good()) n = static_cast<unsigned short>(x);
	return *this;
}


iXDR_Stream& 
iXDR_Stream::operator>>(int& n)
{
	sentry qx(*this);
	if (not qx) return *this;

	long x;
	*this >> x;
	if (good()) n = x;
	return *this;
}


iXDR_Stream& 
iXDR_Stream::operator>>(unsigned int& n)
{
	sentry qx(*this);
	if (not qx) return *this;

	unsigned long x;
	*this >> x;
	if (good()) n = x;
	return *this;
}


iXDR_Stream& 
iXDR_Stream::operator>>(long& n)
{
	sentry qx(*this);
	if (not qx) return *this;

	XDR_Char c;
	get(c);
	if (good()) {
#if BYTE_ORDER == LITTLE_ENDIAN
		c.swap_bytes();
#endif
		::memcpy(&n, c.data, sizeof(n));
	}
	return *this;
}


iXDR_Stream& 
iXDR_Stream::operator>>(unsigned long& n)
{
	sentry qx(*this);
	if (not qx) return *this;
	
	XDR_Char c;
	get(c);
	if (good()) {
#if BYTE_ORDER == LITTLE_ENDIAN
		c.swap_bytes();
#endif
		::memcpy(&n, c.data, sizeof(n));
	}
	return *this;
}


iXDR_Stream& 
iXDR_Stream::operator>>(float& f)
{
	sentry qx(*this);
	if (not qx) return *this;
	
	XDR_Char c;
	get(c);
	if (good()) {
#if BYTE_ORDER == LITTLE_ENDIAN
		c.swap_bytes();
#endif
		::memcpy(&f, c.data, sizeof(f));
	}
	return *this;
}


iXDR_Stream& 
iXDR_Stream::operator>>(double& f)
{
	sentry qx(*this);
	if (not qx) return *this;
	
	XDR_Char c[2];
	read(c, 2);
	if (good()) {
#if BYTE_ORDER == LITTLE_ENDIAN
		c[0].swap_bytes();
		c[1].swap_bytes();
		std::swap(c[0], c[1]);
#endif
		::memcpy(&f, c, sizeof(f));
	}
	return *this;
}


/*** TBD
iXDR_Stream&
iXDR_Stream::operator>>(long double& f);
***/


iXDR_Stream&
iXDR_Stream::operator>>(XDR_Streambuf* sb)
{
	sentry qx(*this);
	if (not qx) return *this;
	
	if (not sb) {
		setstate(failbit);
		return *this;
	}
	
	for (;;) {
		int_type x = rdbuf()->sgetc();
		if (x == traits_type::eof()) {
			setstate(eofbit);
			break;
		}
		int_type y = sb->sputc(traits_type::to_char_type(x));
		if (y == traits_type::eof()) break;
		rdbuf()->sbumpc();
	}
	return *this;
}


size_t
iXDR_Stream::get_opaque(void* d, size_t l)
{
	sentry qx(*this);
	if (not qx) return 0;

	unsigned char* data = static_cast<unsigned char*>(d);
	XDR_Char* xdata = reinterpret_cast<XDR_Char*>(data);
	size_t k = l / 4;
	read(xdata, k);
	if (not good()) return _gcount * 4;
	l    -= k * 4;
	data += k * 4;
	if (l > 0) {
		XDR_Char c;
		get(c);
		if (good()) {
			::memcpy(data, c.data, l);
			data += l;
		}
	}
	return data - static_cast<unsigned char*>(d);
}


size_t
iXDR_Stream::vget_opaque(void* d, size_t l)
{
	unsigned long len;
	*this >> len;
	if (good() and len <= l) return get_opaque(d, len);
	if (good()) setstate(failbit);
	return 0;
}


size_t
iXDR_Stream::vget_string(char* s, size_t l)
{
	size_t len = vget_opaque(s, l);
#ifdef __370__
	if (good()) std::transform(s, s+len, s, ASCII2EBCDIC());
#endif
	return len;
}


iXDR_Stream::int_type
iXDR_Stream::get()
{
	sentry qx(*this);
	if (not qx) return traits_type::eof();
	
	_gcount = 0;
	int_type x = rdbuf()->sbumpc();
	if (x == traits_type::eof()) {
		setstate(badbit);
	} else {
		++_gcount;
	}
	return x;
}


iXDR_Stream& 
iXDR_Stream::get(char_type& c)
{
	int_type x = get();
	if (good()) c = traits_type::to_char_type(x);
	return *this;
}


iXDR_Stream& 
iXDR_Stream::read(char_type* buf, std::streamsize n)
{
	sentry qx(*this);
	if (not qx) return *this;
	
	_gcount = 0;
	while (n > 0) {
		get(*buf);
		if (not good()) break;
		++buf; ++_gcount;
		--n;
	}
	return *this;
}


int
iXDR_Stream::sync()
{
	if (not rdbuf()) return  -1;
	int_type x = rdbuf()->pubsync();
	if (x == traits_type::eof()) {
		setstate(badbit);
		return x;
	}
	return 0;
}


iXDR_Stream::pos_type
iXDR_Stream::tellg()
{
	if (fail()) return pos_type(-1);
	return rdbuf()->pubseekoff(0, cur, in);
}


iXDR_Stream& 
iXDR_Stream::seekg(pos_type pos)
{
	if (not fail()) rdbuf()->pubseekpos(pos);
	return *this;
}


iXDR_Stream& 
iXDR_Stream::seekg(off_type off, std::ios_base::seekdir dir)
{
	if (not fail()) rdbuf()->pubseekoff(off, dir);
	return *this;
}


//
// iXDR_Stream -
//	non-member functions
iXDR_Stream&
operator>>(iXDR_Stream& ixs, std::basic_string<unsigned char>& str)
{
	iXDR_Stream::sentry qx(ixs);
	if (not qx) return ixs;

	str.erase();
	unsigned long len;
	ixs >> len;
	if (ixs.good()) {
		while (len > 0) {
			XDR_Char c;
			ixs.get(c);
			if (not ixs.good()) break;
			int k = (len > 4) ? 4 : len;
			for (int i = 0; i < k; ++i) {
				str += c.data[i];
			}
			len -= k;
		}
	}
	return ixs;
}


iXDR_Stream& 
operator>>(iXDR_Stream& ixs, char* s)
{
	iXDR_Stream::sentry qx(ixs);
	if (not qx) return ixs;
	
	long len;
	ixs >> len;
	if (ixs.good()) {
		if (ixs.width() > 0 and len >= ixs.width()) {
			ixs.width(0);
			ixs.setstate(std::ios_base::failbit);
			return ixs;
		}
		ixs.width(0);
		ixs.get_opaque(reinterpret_cast<unsigned char*>(s), len);
		s[len] = '\0';
#ifdef __370__
		std::transform(s, s+len, s, ASCII2EBCDIC());
#endif
	}
	return ixs;
}


iXDR_Stream& 
operator>>(iXDR_Stream& ixs, std::string& str)
{
	iXDR_Stream::sentry qx(ixs);
	if (not qx) return ixs;
	
	str.erase();
	long len;
	ixs >> len;
	if (ixs.good()) {
		if (ixs.width() > 0 and len > ixs.width()) {
			ixs.width(0);
			ixs.setstate(std::ios_base::failbit);
			return ixs;
		}
		ixs.width(0);
		while (len > 0) {
			XDR_Char c;
			ixs.get(c);
			if (not ixs.good()) break;
			int k = (len > 4) ? 4 : len;
			for (int i = 0; i < k; ++i)
				str += c.data[i];
			len -= k;
		}
#ifdef __370__
		std::transform(str.begin(), str.end(), str.begin(), ASCII2EBCDIC());
#endif
	}
	return ixs;
}


std::pair<unsigned char*, size_t>
vget_opaque(iXDR_Stream& ixs, size_t max)
{
	std::pair<unsigned char*, size_t> nothing(0, 0);
	iXDR_Stream::sentry qx(ixs);
	if (not qx) return nothing;

	long l;
	ixs >> l;
	if (ixs.fail() or l == 0) return nothing;
	if (size_t(l) > max) {
		ixs.setstate(std::ios_base::failbit);
		return nothing;
	}

	unsigned char* p = new unsigned char[l];
	ixs.get_opaque(p, l);
	if (ixs.fail()) return nothing;
	return std::make_pair(p, l);
}


std::pair<char*, size_t>
vget_string(iXDR_Stream& ixs, size_t max)
{
	std::pair<char*, size_t> nothing(new char[1],0);
	nothing.first[0] = '\0';
	iXDR_Stream::sentry qx(ixs);
	if (not qx) return nothing;

	long l;
	ixs >> l;
	if (ixs.fail()) return nothing;
	if (size_t(l) > max) {
		ixs.setstate(std::ios_base::failbit);
		return nothing;
	}

	char* p = new char[l+1];
	ixs.get_opaque(p, l);
	if (ixs.fail()) {
		delete[] p;
		return nothing;
	}
	p[l] = '\0';
#ifdef __370__
	std::transform(p, p+l, p, ASCII2EBCDIC());
#endif
	delete[] nothing.first;
	return std::make_pair(p, l);
}


}  //> namespace Util
