/*******************************************************************************
  File: XDRStream_writer.cpp
  Copyright (c) 1999 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, to prevent the 
  continued distribution of this Software under an open source distribution
  model. This includes, but is not limited to, such things as copyrighting
  or claiming authorship of the Software.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
$Header:$
  Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
  Implements the oXDR_Stream and the XDR_Stream.
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/
// My header
#include "XDRStream.h"

// String conversion
#include "E2A.h"

namespace Util {


oXDR_Stream::oXDR_Stream()
{
	// empty
}


oXDR_Stream::oXDR_Stream(XDR_Streambuf* sb)
{
	init(sb);
}


oXDR_Stream::~oXDR_Stream()
{
	// empty
}


oXDR_Stream& 
oXDR_Stream::operator<<(long n)
{
	sentry qx(*this);
	if (not qx) return *this;

	XDR_Char c;
	::memcpy(c.data, &n, sizeof(c.data));
#if BYTE_ORDER == LITTLE_ENDIAN
	c.swap_bytes();
#endif
	put(c);
	return *this;
}


oXDR_Stream& 
oXDR_Stream::operator<<(unsigned long n)
{
	sentry qx(*this);
	if (not qx) return *this;

	XDR_Char c;
	::memcpy(c.data, &n, sizeof(c.data));
#if BYTE_ORDER == LITTLE_ENDIAN
	c.swap_bytes();
#endif
	put(c);
	return *this;
}


oXDR_Stream& 
oXDR_Stream::operator<<(float f)
{
	sentry qx(*this);
	if (not qx) return *this;

	XDR_Char c;
	::memcpy(c.data, &f, sizeof(c.data));
#if BYTE_ORDER == LITTLE_ENDIAN
	c.swap_bytes();
#endif
	put(c);
	return *this;
}


oXDR_Stream& 
oXDR_Stream::operator<<(double f)
{
	sentry qx(*this);
	if (not qx) return *this;

	XDR_Char c[2];
	::memcpy(c, &f, sizeof(f));
#if BYTE_ORDER == LITTLE_ENDIAN
	std::swap(c[0], c[1]);
	c[0].swap_bytes();
	c[1].swap_bytes();
#endif
	write(c, 2);
	return *this;
}

/***
oXDRStream&
oXDR_Stream::operator<<(long double f);   TBD
***/


// copy another streambuf into this stream
oXDR_Stream& 
oXDR_Stream::operator<<(XDR_Streambuf* sb)
{
	sentry qx(*this);
	if (not qx) return *this;
	
	if (not sb) {
		setstate(badbit);
		return *this;
	}

	size_t count = 0;
	for (;;) {
		int_type x = sb->sgetc();
		if (x == traits_type::eof()) {
			if (count == 0) setstate(failbit);
			break;
		}
		put(traits_type::to_char_type(x));
		if (not good()) break;
		++count;
		sb->sbumpc();
	}
	return *this;
}
	

// opaque data inserters
oXDR_Stream& 
oXDR_Stream::put_opaque(const void* d, size_t l)
{
	sentry qx(*this);
	if (not qx or l == 0) return *this;

	const unsigned char* data = reinterpret_cast<const unsigned char*>(d);
	const char_type* xdata = reinterpret_cast<const char_type*>(data);
	std::streamsize k = l / 4;
	write(xdata, k);
	if (good()) {
		l    -= k * 4;
		data += k * 4;		
		if (l > 0) {
			XDR_Char c;
			::memcpy(c.data, data, l);
			c.pad(4 - l);
			put(c);
		}
	}
	return *this;
}


oXDR_Stream& 
oXDR_Stream::vput_opaque(const void* d, size_t l)
{
	*this << static_cast<unsigned long>(l);
	return put_opaque(d, l);
}


oXDR_Stream&
oXDR_Stream::vput_string(const char* str, size_t l)
{
	sentry qx(*this);
	if (not qx) return *this;
	if (l == 0) return *this << static_cast<long>(l);

#ifdef __370__
	char* tstr = new char[l];
	std::transform(tstr, tstr+l, tstr, ASCII2EBCDIC());
	vput_opaque(tstr, l);
	delete[] tstr;
#else
	vput_opaque(str, l);
#endif
	return *this;
}


oXDR_Stream& 
oXDR_Stream::put(char_type c)
{
	int_type x = rdbuf()->sputc(c);
	if (x == traits_type::eof()) setstate(badbit);
	return *this;
}


oXDR_Stream& 
oXDR_Stream::write(const char_type* s, std::streamsize n)
{
	std::streamsize x = rdbuf()->sputn(s, n);
	if (x != n) setstate(badbit);
	return *this;
}


oXDR_Stream& 
oXDR_Stream::flush()
{
	sentry qx(*this);
	if (not qx) return *this;
	
	int x = rdbuf()->pubsync();
	if (x == -1) setstate(badbit);
	return *this;
}


oXDR_Stream::pos_type 
oXDR_Stream::tellp()
{
	if (fail()) return pos_type(-1);
	return rdbuf()->pubseekoff(0, cur, out);
}


oXDR_Stream& 
oXDR_Stream::seekp(pos_type pos)
{
	if (not fail())	rdbuf()->pubseekpos(pos);
	return *this;
}


oXDR_Stream& 
oXDR_Stream::seekp(off_type off, ios_base::seekdir dir)
{
	if (not fail()) rdbuf()->pubseekoff(off, dir);
	return *this;
}


//
// XDR_Stream
//
XDR_Stream::XDR_Stream()
{
	// empty
}


XDR_Stream::XDR_Stream(XDR_Streambuf* sb)
{
	init(sb);
}


XDR_Stream::~XDR_Stream()
{
	// empty
}


} //> namespace Util
